/*
   This program is based on Figure 7.4, page 217, from
   "Principles of Parallel Programming" by Lin & Snyder.

   This program divides p processes into k groups (where
   we assume that k divides p without remainder).
   The number of groups, k, is a command line parameter.

   This program uses MPI_Comm_group(), MPI_Group_incl(),
   and MPI_Comm_create() to create the group communicators.
   You can also use MPI_Comm_split() to create the communicators,
   which is sometimes easier to use.
*/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "mpi.h"
#include <windows.h>  // used for Sleep() only

int main(int argc, char** argv)
{
   int p;           /* Number of processes */
   int k;           /* Number of process groups */
   int groupSize;   /* Number of processes in a group */
   int rank;        /* Rank of this process */
   int groupNumber; /* Number for the group that this process is a memeber of */
   int groupRank;   /* The "rank" of this process within its group */
   int myGroupRank; /* The official rank of this process within its group */
   int myGroupSize; /* The official size of this process's group */

   int* ranks;
   MPI_Group ourGroup;
   MPI_Comm groupComm;
   MPI_Group globalGroup;

   int random, workTime, i;

   MPI_Init(&argc, &argv);
   MPI_Comm_size(MPI_COMM_WORLD, &p);
   MPI_Comm_rank(MPI_COMM_WORLD, &rank);

   k = atoi( argv[1] );
   groupSize = p / k;
   groupNumber = rank / groupSize;
   groupRank   = rank % groupSize;

   /* Initialize the ranks[] array */
   ranks = (int*)malloc( p * sizeof(int) );
   for (i = 0; i < p; i++)
   {
      ranks[i] = i;
   }

   /* Extract the original group hsndle */
   MPI_Comm_group(MPI_COMM_WORLD, &globalGroup);

   /* Define the new group */
   MPI_Group_incl(globalGroup, groupSize, ranks+(groupNumber*groupSize), &ourGroup);

   /* Define the new communicator */
   MPI_Comm_create(MPI_COMM_WORLD, ourGroup, &groupComm);

   /* What is my rank within my group communicator? */
   MPI_Group_rank(ourGroup, &myGroupRank);
   /* What is the size of my group communicator? */
   MPI_Group_size(ourGroup, &myGroupSize);

   /* Identify the console window */
   printf("I'm rank %d out of %d processes in MPI_COMM_WORLD.\n", rank, p);
   printf("Unofficially, I'm number %d in group %d.\n", groupRank, groupNumber);
   printf("Officially, I'm rank %d out of %d processes in my group communicator.\n", myGroupRank, myGroupSize);

   /* Do some "work" */
   srand(rank*100); /* seed the random number generator */
   workTime = 1+ rand()%10;
   printf("Doing %d seconds of work.\n", workTime);
   for (i = workTime-1; i >=0; i--)
   {
      Sleep(1000);             // Win32
      //sleep(workTime/1000);  // Posix
      printf("      %d\n", i);
   }

   /* Rendezvous with all the other processes in my group */
   /* Broadcast a random number within each group */
   random = rand(); /* generate a random number (only the group root's number will be used) */
   MPI_Bcast(&random, 1, MPI_INT, 0, groupComm); /* sent by the group's root process */
   printf("My group's random number = %d\n", random);

   MPI_Finalize();
   return 0;
}
